module agn_stars

   use star_lib
   use star_def
   use const_def
   use math_lib

   implicit none
   public

   ! Radiation-dominated AGN
   real(dp), parameter :: gamma_agn = 4d0 / 3d0
   ! Indices of angular momentum variables in extras_rpar
   integer, parameter :: i_mu_J = 18
   integer, parameter :: i_var_J = 19
   integer, parameter :: i_J_crit = 20
   ! For time-step control
   real(dp), parameter :: max_delta_ln_m = 1d-1
   real(dp), parameter :: blend_safety = 1d-3
   real(dp), parameter :: burn_safety = 3d-1

   real(dp) :: const_rho               ! Ambient density in g/cm^3
   real(dp) :: const_Tb                ! Ambient temperature in K
   real(dp) :: const_csb               ! Ambient sound speed in cm/s, corresponds to const_Tb
   real(dp) :: const_kappa             ! Average opacity of accretion stream (cm^2/s) 
   real(dp) :: atm_blend_time          ! Time (yrs) over which star falls into AGN (used for blending atmospheric boundary condition)
   real(dp) :: mdot_blend_time         ! Time (yrs) over which star falls into AGN (used for blending atmospheric boundary condition)
   real(dp) :: min_Teff                ! Extra term added to Teff^4, phased out during the min_Teff_blend_time.
   real(dp) :: min_Teff_blend_time     ! Time over which to phase out min_Teff
   real(dp) :: Omega_AGN               ! AGN angular frequency
   real(dp) :: Mass_AGN                ! SMBH mass
   real(dp) :: last_gain               ! mass gain in last time step
   real(dp) :: atm_blend = 1d-20         ! Current state of the blend
   real(dp) :: mdot_blend = 1d-20       ! Current state of the blend
   real(dp) :: lost_mass(200)          ! Tracks lost mass

   contains


   subroutine my_other_D_mix(id, ierr)
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      integer :: k, nz
      real(dp) :: Ledd, Leff, Dmix, R

      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)
      nz = s%nz

      Leff = s% job% extras_rpar(3) ! prev_luminosity
      R = s% job% extras_rpar(4) ! prev_radius
      Ledd = 3.2d4 * s%m(1) * Lsun / Msun
      do k=1,nz
         Dmix = s%scale_height(k) * pow(Leff / (4 * pi * pow2(R) * s%rho(k)), 1d0 / 3d0)
         s%D_mix(k) = s%D_mix(k) + Dmix * tanh(pow(Leff / Ledd, s% x_ctrl(5)))  ! x_ctrl(5) = Power of gamma for increasing Mixing
      end do

      ! Save variables for history

      s% xtra1_array(3) = tanh(pow(Leff / Ledd, s% x_ctrl(5))) ! Extra mixing factor (constant through the star) induced by accretion

   end subroutine my_other_D_mix

   integer function my_other_timestep_limit( &
      id, skip_hard_limit, dt, dt_limit_ratio)
      use const_def, only: dp
      integer, intent(in) :: id
      logical, intent(in) :: skip_hard_limit
      real(dp), intent(in) :: dt
      real(dp), intent(inout) :: dt_limit_ratio
      type (star_info), pointer :: s
      integer :: j
      real(dp) :: eps_nuc, burn_time, mix_time, f_mix, min_burn_time, min_eps_nuc, min_f_mix, diff_integral
      integer :: ierr
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return

      min_burn_time = 1d100
      diff_integral = 0d0
      do j=2,s%nz
         if (s%m(1) - s%m(j) > last_gain) then
            ! Tracks the diffusion front, assuming a constant source at the surface
            diff_integral = diff_integral + (s%r(1)-s%r(j)) * s%dm(j) / (4d0 * pi * pow2(s%r(j)) * s%rho(j) * s%D_mix(j)) 
            mix_time = diff_integral

            if (mix_time / s%dt < 20d0) then
               f_mix = min(1d0, exp(-mix_time / s%dt))
            else
               f_mix = 1d-50
            end if

            eps_nuc = s%eps_nuc(j)
            burn_time = pow2(s%csound(j)) / (f_mix * eps_nuc) ! (thermal energy / burning energy generation)

            if (burn_time < min_burn_time) then
               min_burn_time = burn_time
               min_eps_nuc = eps_nuc
               min_f_mix = f_mix
            end if
         end if
      end do

      dt_limit_ratio = s%dt / (min_burn_time * burn_safety)
      !write(*,*) 'Burn dt limit:', s%dt / (min_burn_time * burn_safety)

      dt_limit_ratio = max(dt_limit_ratio, abs(s%mstar_dot) * s%dt / (max_delta_ln_m * s%m(1)))
      !write(*,*) 'Mdot dt limit:', abs(s%mstar_dot) * s%dt / (max_delta_ln_m * s%m(1))

      if (s%time / secyer - 1d5 < 1d1*atm_blend_time + 1d1*mdot_blend_time) then
         dt_limit_ratio = max(dt_limit_ratio, s%dt * atm_blend*(1d0-atm_blend) / (atm_blend_time * secyer * blend_safety))
         !write(*,*) 'atm blend limit:', s%dt / (atm_blend_time * secyer * blend_safety)

         dt_limit_ratio = max(dt_limit_ratio, s%dt * mdot_blend*(1d0-mdot_blend) / (mdot_blend_time * secyer * blend_safety))
         !write(*,*) 'mdot blend limit:', s%dt / (mdot_blend_time * secyer * blend_safety)
      end if
      my_other_timestep_limit = keep_going
   end function my_other_timestep_limit

   subroutine set_blend(id, ierr)
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      real(dp) :: initial_age, mdot_start_time

      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)

      initial_age = 1d5

      ! 1d5 is because the starting model is 1d5 years old.
      atm_blend = 0.5d0 * (1d0 + tanh((s%time / secyer - initial_age - atm_blend_time) / (3d-1 * atm_blend_time))) + 1d-20
      mdot_start_time = 3d0 * atm_blend_time
      mdot_blend = 0.5d0 * (1d0 + tanh((s%time / secyer - initial_age - mdot_start_time) / (3d-1 * mdot_blend_time))) + 1d-20

      !write(*,*) 'atm blend, mdot blend:', atm_blend, mdot_blend

   end subroutine set_blend

end module agn_stars
