module atmosphere

   use auto_diff
   use star_lib
   use star_def
   use const_def
   use math_lib
   use agn_stars
   use mass_changes
   use utils_lib

   implicit none

   public
   contains

   subroutine other_surface_PT(id, skip_partials, &
            Teff, lnT_surf, dlnT_dL, dlnT_dlnR, dlnT_dlnM, dlnT_dlnkap, &
            lnP_surf, dlnP_dL, dlnP_dlnR, dlnP_dlnM, dlnP_dlnkap, ierr)

         integer, intent(in) :: id
         logical, intent(in) :: skip_partials
         real(dp), intent(out) :: Teff, &
            lnT_surf, dlnT_dL, dlnT_dlnR, dlnT_dlnM, dlnT_dlnkap, &
            lnP_surf, dlnP_dL, dlnP_dlnR, dlnP_dlnM, dlnP_dlnkap
         integer, intent(out) :: ierr

      logical, parameter :: dbg = .false.
      type (star_info), pointer :: s
      real(dp) :: minT, blend_minT
      type(auto_diff_real_4var_order1) :: M, R, L, kappa
      type(auto_diff_real_4var_order1) :: R_ph, R_bondi, vesc2, Mdot, loss, super_eddington_factor, L_incident
      type(auto_diff_real_4var_order1) :: H_disk, erf_arg, erf_f1, erf_f2, vortpar, R_Hill, tidepar
      type(auto_diff_real_4var_order1) :: L_stream, L_shock, Ledd, T_surf, P_surf, P_rad, P_acc, P_outflow, P_simple
      type(auto_diff_real_4var_order1) :: simple_Teff, simple_lnT, simple_lnP, agn_Teff, agn_lnT, agn_lnP
      type(auto_diff_real_4var_order1) :: lnT, lnP, overall_Teff

      ! Basic setup
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return

      ! Setup autodiff variables
      M = s%m(1)
      M%d1val1 = 1d0

      R = s%r(1)
      R%d1val2 = 1d0

      L = s%L(1)
      L%d1val3 = 1d0

      kappa = const_kappa
      kappa%d1val4 = 0d0

      ! Basic quantities

      ! Bondi radius
      R_bondi = 2d0 * standard_cgrav * s%m(1) / pow2(const_csb)

      ! Mdot
      Mdot = pi * pow2(R_bondi) * const_rho * const_csb
      ! Adjust gain down due to radiation pressure
      Ledd = 3.2d4 * M * Lsun / Msun

      if (s% x_logical_ctrl(1)) then
        R_bondi = R_bondi * (1d0 - L/Ledd)
        R_bondi = max(R_bondi, R)
        Mdot = pi * pow2(R_bondi) * const_rho * const_csb
      else
        Mdot = Mdot * (1d0 - tanh(abs(L / Ledd)))
      end if

      R_ph = pow2(kappa * const_rho) * pow3(R_bondi) ! Photosphere radius
      R_ph = R_ph + R
      R_ph = min(R_ph, R_bondi)

      if (s% x_logical_ctrl(3)) then
         H_disk = pow(2d0, 0.5d0) * const_csb / Omega_AGN
         erf_arg = -R_bondi / H_disk
         erf_f1 = (-2d0/pow(pi, 0.5d0) )*pow( 1d0 - exp(-erf_arg*erf_arg), 0.5d0)
         erf_f2 = pow(pi, 0.5d0)*0.5d0 + (3.1d1/2d2)*exp(-erf_arg*erf_arg) - (3.41d2/8d3)*exp(-2d0*erf_arg*erf_arg)
         Mdot = 0.5d0 * Mdot * pow( pi, 0.5d0 ) * erf_f1 * erf_f2 / erf_arg
         ! Adjust accretion rate to take into account changes in disk density vertically.
         ! Comes from averaging exp(-(z/H)^2) over the surface of a sphere with r = r_Bondi
         ! ERF is approximated by a Burmann series using c1 = 31/200 and c2 = -341/8000
      end if

      if (s% x_logical_ctrl(4)) then
            vortpar = (R_bondi * Omega_AGN)/(12d0 * const_csb) !avg vorticity parameter at R_bondi
            vortpar = (2d0 / (pi * vortpar))*asinh( pow(2d0*vortpar, 1d0/3d0))  !Krumholz, McKee, Klein 2005 eqn. A7
            Mdot = Mdot * min(1d0, vortpar) !pick minimum factor
            ! Adjust accretion rate to take into account shear
      end if

      if (s% x_logical_ctrl(5)) then
        if (s% x_logical_ctrl(1)) then
            R_Hill = pow(standard_cgrav * s%m(1) * (1d0 - L/Ledd)/ (3d0 * Omega_AGN*Omega_AGN), 1d0/3d0)
        else
            R_Hill = pow(standard_cgrav * s%m(1) / (3d0 * Omega_AGN*Omega_AGN), 1d0/3d0)
        end if
            tidepar = pow(R_Hill / R_bondi, 2d0)
            Mdot = Mdot * min(1d0, tidepar) !pick minimum factor

            ! Use min(Bondi, Hill) for accretion rate. Similar to Rosenthal et al. 2020
      end if

      ! Calculate super-eddington mass loss
      vesc2 = 2d0*standard_cgrav*M / R
      super_eddington_factor = 0.1d0
      loss = super_eddington_mdot_autodiff(M,L,R,super_eddington_factor)

      ! Set a minimum on Teff.
      minT = const_Tb

      ! Various luminosities
      L_shock = Mdot * M * standard_cgrav / R
      L_stream = L + L_shock
      L_incident = 4 * pi * pow2(R_ph) * boltz_sigma * pow4(minT)
      L_stream = L_stream + L_incident
      L_stream = max(L_stream, L_incident) ! Avoid NaN's due to bad near-surface grad(T)

      ! Evaluate simple photosphere
      simple_Teff = pow(L_stream / (4d0 * pi * R**2 * boltz_sigma), 0.25d0)
      simple_lnP = log(standard_cgrav * M / (R**2 * kappa))
      P_simple = exp(simple_lnP)
      simple_lnT = log(simple_Teff) ! First cell is tau=2/3 in this simple solution.

      ! Evaluate complicated photosphere
      agn_Teff = pow(L_stream / (4 * pi * boltz_sigma * pow2(R_ph)), 0.25d0)
      T_surf = agn_Teff * pow(R_ph / R, 5d0 / 8d0)
      agn_lnT = log(T_surf)

      ! P_surf
      P_acc = const_rho * pow2(const_csb) * pow(R_bondi / R, 2.5d0)
      P_rad = crad * pow4(T_surf) / 3d0
      P_outflow = - loss * pow(vesc2, 0.5d0) / (4 * pi * pow2(R))
      P_surf = P_acc + P_rad + P_outflow + P_simple
      agn_lnP = log(P_surf)

      ! blend gets set in extras_start_step
      lnT = agn_lnT * atm_blend + (1d0 - atm_blend) * simple_lnT
      lnP = agn_lnP * atm_blend + (1d0 - atm_blend) * simple_lnP
      overall_Teff = agn_Teff * atm_blend + (1d0 - atm_blend) * simple_Teff

      ! Format for output
      lnT_surf = lnT%val
      dlnT_dlnM = lnT%d1val1 * M%val
      dlnT_dlnR = lnT%d1val2 * R%val
      dlnT_dL = lnT%d1val3
      dlnT_dlnkap = lnT%d1val4 * kappa%val

      lnP_surf = lnP%val
      dlnP_dlnM = lnP%d1val1 * M%val
      dlnP_dlnR = lnP%d1val2 * R%val
      dlnP_dL = lnP%d1val3
      dlnP_dlnkap = lnP%d1val4 * kappa%val

      Teff = overall_Teff%val

      if (is_bad(lnT_surf) .or. is_bad(lnP_surf) .or. is_bad(Teff) .or. dbg) then
         write(*,*) 'Bad ATM'
         write(*,*) '----------- Inputs -----------'
         write(*,*) 'M',M
         write(*,*) 'R',R
         write(*,*) 'L',L
         write(*,*) 'kap',kappa
         write(*,*) 'Mdot', Mdot
         write(*,*) '----------- Derived -----------'
         write(*,*) 'Rph', R_ph
         write(*,*) 'R_Bondi', R_bondi
         write(*,*) 'L_stream', L_stream
         write(*,*) 'Blend', atm_blend
         write(*,*) '----------- Outputs -----------'
         write(*,*) 'lnT',lnT
         write(*,*) 'lnP',lnP
         write(*,*) 'Teff',overall_Teff
      end if

      s% job% extras_rpar(3) = L%val
      s% job% extras_rpar(4) = R%val

      ! Save variables for history

      s% xtra1_array(4) = L_shock%val    ! Accretion Shock Luminosity
      s% xtra1_array(5) = L_stream%val   ! Stream Luminosity
      s% xtra1_array(6) = R_ph%val       ! Photospheric Radius
      s% xtra1_array(7) = R_bondi%val    ! R Bondi
      s% xtra1_array(10) = R%val         ! R Shock (R 'Surface' of MESA model)
      s% xtra1_array(11) = lnT%val       ! log2 T Shock (T 'Surface' of MESA model)

   end subroutine other_surface_PT



end module atmosphere
